"""Handling of quaternion, vector, and other 3d object data in Python.

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    scalar
    vector
    quaternion

"""

__name__ = "orix"
__version__ = "0.1.0"
__author__ = "Ben Martineau"
__author_email__ = "pyxem.team@gmail.com"
__description__ = "Orientation, rotation, quaternion, and crystal symmetry handling in Python."
